(* Scanner.mll - Performs lexical analysis to generate tokens *)

{ open Parser }

let digit = ['0' - '9']
let ip = digit+   				(* Float numbers - Pre-decimal point *)
let fp = digit+						(* Float numbers - post-decimal point*)
let ie = digit+						(* Integer numbers *)
let dp = '.' 
let exp = 'e' ('+'|'-')? ip
(* Clearly ip, fp, ie are all identical, but for clarity we separate *)

rule token = parse
  [' ' '\t' '\r' '\n'] { token lexbuf } (* Whitespace *)
| "#"      { comment lexbuf }           (* Comments *)
| ';'      { SEMI }											(* Punctuation *)
| '('      { LPAREN }
| ')'      { RPAREN }
| '{'      { LBRACE }
| '}'      { RBRACE }
| ','      { COMMA }
| '+'      { PLUS }											(* Arithmetic *)
| '-'      { MINUS }
| '*'      { TIMES }
| '/'      { DIVIDE }
| '%'	   	 { MOD }
| '='      { ASSIGN }										(* Assignment *)
| "=="     { EQ }												(* Comparison Logic *)
| "!="     { NEQ }
| '<'      { LT }
| "<="     { LEQ }
| ">"      { GT }
| ">="     { GEQ }
| "&&"     { AND }											(* Boolean Logic *)
| "!"      { NOT }
| "||"     { OR }
| "return" { RETURN }										(* Control Flow*)
| "if"     { IF }
| "else"   { ELSE }
| "for"    { FOR }
| "while"  { WHILE }
| "break"  { BREAK }
| "int"    { INT }											(* Data Types *)
| "float"  { FLOAT }
| "bool"   { BOOL }
| "char"   { CHAR }
| "string" { STRING }
| "NULL"   { NULL }											(* Null *)
| "void"   { VOID }                     (* Void -- only used as return type *)
| '['      { LSQPAREN }           			(*TaML exclusive*)
| ']'      { RSQPAREN }           
| '@'      { WILDCARD }           
| '~'      { RANGE }              
| "cell"   { CELL }               
| "line"   { LINE }               
| "table"  { TABLE }              
| "func"   { FUNC }								
| "print"  { PRINT }              
| "^"      { REF }											(* Indicates an ID is for a Cell type. Example: ^myCell = 500; *)
(* Literals/Constants *)
| ie as lit { INTLITERAL(int_of_string lit) }
| (((ip dp fp)|(ip dp)|(dp fp))(exp)?)|(ip exp) as lit { FLOATLITERAL(float_of_string lit) }
| ['\"'] ([^ '\"']* as lit) ['\"'] { STRINGLITERAL(lit) }				(* String can accept anything but " or /".  Can be empty *)
| ['\''] ( _ as lit) ['\''] { CHARLITERAL(lit) }								(* Char can accept anything. Cannot be empty *)
| "true" | "false" as lit { BOOLLITERAL(bool_of_string lit) }

(* Variables *)
| ['a'-'z' 'A'-'Z']['a'-'z' 'A'-'Z' '0'-'9' '_']* as ident { ID(ident) }

(* End of File *)
| eof { EOF }

(* Anything else*)
| _ as char { raise (Failure("illegal character " ^ Char.escaped char)) }



and comment = parse
  ['\r' '\n'] { token lexbuf }    (* Return to normal scanning *)
| _    { comment lexbuf }         (* Ignore other characters *)
