//////////////////////////////////////////////////////////////////////////////////
// 	Line Type - 1d array composed of Cells,	   									//
//  where the cells reference cells in a table 									//
// NOTE: if the LINE is holding floats, all numbers must end with				//
// an "f".  For example, mylINE.setCellValue(1, 3.5f)							//
// This is because Java automatically assumes decimal point numbers				//
// are double unless explicitly case to float -- resulting in compile errors	//
//////////////////////////////////////////////////////////////////////////////////


package java_lib;

//E must be Integer or Float
public class Line<E> {
	private Cell<E>[] line;
	private int lineLength;

	
	//////////////////////////////////////////////////////////////////
	// Initialize to null.  										//
	// If the line is declared and initialized to a table			//
	// in the same line (e.g., line l = myTable[@,3] ) just 		//
	// call assignLine directly after								//
	//////////////////////////////////////////////////////////////////
	public Line(){
		line = null;
		lineLength=0;
	}
	
	/////////////////////////////////////////////////////////////////
	// Create a line that isn't assigned to a table -- it's cells  //
	// get their own spot in memory								   //
	/////////////////////////////////////////////////////////////////
	public Line(int length){
		lineLength = length;
		line = new Cell[length];
		for (int i=0; i<length; i++){
				Cell<E> c = new Cell<E>();
				line[i] = c;
		}
	}
	
	
	//==============================================================//
	//					LINE ASSIGNMENT (TO TABLE)			   		//
	//	Assign the line to a contiguous row of cells in the table	//
	//==============================================================//
	
	//////////////////////////////////////////////////////////////////
	// Index Syntax: [1~4, 5] or [3, 4~5]							//
	//////////////////////////////////////////////////////////////////
	public void assignLine(Table<E> t, int startRow, int endRow, int startCol, int endCol){
		//Error Checking
		if(endRow<startRow || endCol<startCol )
			throw new IllegalArgumentException("Error start row/column must be less than end row/column");	
		
		// when we are creating a "row" line
		if(startRow == endRow ){
			lineLength = endCol - startCol + 1; 
			line = new Cell[lineLength];
			for (int i=0; i<lineLength; i++)
				line[i] = t.getCell(startRow,startCol + i);
		}
		
		// when we are creating a "column" line 
		else if (startCol == endCol){
			lineLength = endRow - startRow + 1;
			line = new Cell[lineLength];
			for (int i=0; i<lineLength; i++)
				line[i] = t.getCell(startRow + i, startCol);
		}	
			
		else
			throw new IllegalArgumentException("Error: when assigning lines, either the startRow/endRow or startCol/endCol MUST be equal");
	}
	
	//////////////////////////////////////////////////////////////////
	// Index Syntax: [@, 5] 										//
	//////////////////////////////////////////////////////////////////
	public void assignLine(Table<E> t, String a1, String a2, int startCol, int endCol){
		//Error Checking
		if(!a1.equals("ALL") || !a2.equals("ALL"))
			throw new IllegalArgumentException("Error -- both rows are not @");	
		if(endCol != startCol )
			throw new IllegalArgumentException("Error: end column and start column must be equal when row is @");	
		
		lineLength = t.numRows();
		line = new Cell[lineLength];
		for (int i=0; i<lineLength; i++)
			line[i] = t.getCell(i, startCol);

	}
	
	//////////////////////////////////////////////////////////////////
	// Index Syntax: [5, @] 										//
	//////////////////////////////////////////////////////////////////
	public void assignLine(Table<E> t, int startRow, int endRow, String a1, String a2){
		//Error Checking
		if(!a1.equals("ALL") || !a2.equals("ALL"))
			throw new IllegalArgumentException("Error -- both columns are not @");	
		if(endRow != startRow )
			throw new IllegalArgumentException("Error: end row and start row must be equal when column is @");	
		
		lineLength = t.numColumns();
		line = new Cell[lineLength];
		for (int i=0; i<lineLength; i++)
			line[i] = t.getCell(startRow, i);

	}
	

	
	//======================================================================//
	//				LINE COPIERS											//
	// Creates & returns a new line that is a COPY of cells in 				//
	// the old line. (e.g., table has new memory allocation)				//
	// Note that the new Line will be of the same type (int or float)  		//
	// as the old line.														//
	//======================================================================//
	
	//////////////////////////////////////////////////////////////////////	
	// Example Usage: Line newL = oldL.createLineCopy(1, 5);			//
	//////////////////////////////////////////////////////////////////////
	public Line<E> createLineCopy(int startIndex, int endIndex ){
		//Error Checking
		if(endIndex < startIndex )
			throw new IllegalArgumentException("Error start index must be less than end index");	
		
		// Initialize the new line
		int newLength = (endIndex - startIndex) + 1;
		Line<E> newLine = new Line<E>(newLength);
		
		// Copy cells into the new table
		copyCells(startIndex, endIndex, newLine);
		return newLine;
	}
	
	
	//////////////////////////////////////////////////////////////////////	
	// Example Usage: Line newL = oldL.createLineCopy("ALL", "ALL");	//
	//////////////////////////////////////////////////////////////////////
	public Line<E> createLineCopy(String a1, String a2 ){
		//Error Checking
		if(!a1.equals("ALL") || !a2.equals("ALL"))
			throw new IllegalArgumentException("Error -- both columns are not @");	
		
		// Initialize the new line
		int newLength = lineLength;
		Line<E> newLine = new Line<E>(newLength);
		
		// Copy cells into the new table
		copyCells(0, newLength-1, newLine);
		return newLine;
	}
	
	
	// HELPER FUNCTION: for copying cells into new table
	private void copyCells(int startIndex, int endIndex, Line<E> newLine) {
		int a=0;				// newTable's index

		for (int i=startIndex; i<=endIndex; i++){
			Cell<E> c = new Cell<E>();
			c.setVal(line[i].getVal());
			newLine.setCell(a,c);
			a++;
		}
	}
	
	//=========================================================//
	//				GETTERS & SETTERS						   //
	//=========================================================//
	public E getCellValue(int index){
		return line[index].getVal();
	}
	

	// Set single cell to a value
	public void setVal(int index, E newValue){
		line[index].setVal(newValue);
	}
    
	
	// Sets entire line to a single Value
	public void setVal(E newValue){
		for(int i=0; i<lineLength; i++)
			line[i].setVal(newValue);
	}
	
	// Sets subrange of line to a single Value
	public void setVal(int startIndex, int endIndex, E newValue){
		if(endIndex < startIndex )
			throw new IllegalArgumentException("Error start index must be less than end index");	

		for(int i=startIndex; i<=endIndex; i++)
			line[i].setVal(newValue);
	}
	
	//////////////////////////////////////////////////////////////////
	// Useful if we want to set a new cell etc. to a cell			//
	// in this line (and it's referenced table cell)  				//
	// e.g., "cell c = myLine[3]" could leverage this function		//
	//////////////////////////////////////////////////////////////////
	public Cell<E> getCell(int index){
		return line[index];
	}
	
	private void setCell(int index, Cell<E> c){
		line[index] = c;
		
	}
	
	
	//=========================================================//
	//				PRINT FUNCTION							   //
	//=========================================================//
	public void print(){
		StringBuffer[] rowText = new StringBuffer[4]; 	// text of each Row
		int colLength;									// max length of a particular column
		String bars = "===============================================";
		
		// Setup Row Values & initialize StringBuffers
		rowText[0] = new StringBuffer("" + '\t');
		rowText[1] = new StringBuffer("" + '\t');
		rowText[2] = new StringBuffer("1" + '\t' + "| ");
		rowText[3] = new StringBuffer("" + '\t');
		
		// Save Data for each line column into rowText[2]
		for (int i=0; i<lineLength; i++){
			
			// Find Max Length in a certain column
			colLength = String.valueOf(line[i].getVal()).length();
			
			// Save Data of the line, setting length of column to fit maxLength
			String curNum = String.valueOf(line[i].getVal());
			String paddedNum = String.format("%1$-" + colLength + "s", curNum);
			rowText[2].append(paddedNum + " | ");
				
			// Save Data for Column Labels
			String curLabel;
			if(i<26)
				curLabel = String.valueOf((char)(i + 65));
			else
				curLabel = String.valueOf((char)(i + 39))+String.valueOf((char)(i + 39));
			String paddedLabel = String.format("%1$-" + colLength + "s", curLabel);
			rowText[0].append("  " + paddedLabel + " ");			
			
			// Save Data for Bottom & Top Bars
			String bottomBar = bars.substring(0, colLength+3); 
			rowText[1].append(bottomBar);
			rowText[3].append(bottomBar);
		}
		
		// Print rows
		for (int i=0; i<=3; i++){
			System.out.println(rowText[i]);
		}	
	}
}
